/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is APNG Edit.
 *
 * The Initial Developer of the Original Code is Mozilla.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Justin Dolske <dolske@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const Cc = Components.classes;
const Ci = Components.interfaces;

var APNGedit = {

    _logService  : null, // Error Console logging
    _ioService   : null, // IO service, .newURI used for string->nsIURI
    _prefs       : null, // "extensions.apngedit." pref branch

    _tmpCanvas : null,
    ani : null, // the <xul:image> showing the animation

    autoPreview : true,

    aniInfo : {
        sizeInitialized : false,
        width     : 0,
        height    : 0,
        playCount : 0,
        skipFirstFrame : false
    },


    /*
     * log
     *
     * Log debug messages to the Javascript console.
     * Note: javascript.options.showInConsole must be enabled
     */
    log : function (message) {
        if (!this._prefs || !this._prefs.getBoolPref("debug")) { return; }
        dump("APNG Edit: " + message + "\n");
        this._logService.logStringMessage("APNG Edit: " + message);
    },


    /*
     * init
     *
     */
    init : function () {
        this._logService = Cc['@mozilla.org/consoleservice;1'].getService();
        this._logService.QueryInterface(Ci.nsIConsoleService);

        this._prefs = Cc["@mozilla.org/preferences-service;1"].getService();
        this._prefs.QueryInterface(Ci.nsIPrefService);
        this._prefs = this._prefs.getBranch("extensions.apngedit.");

        this.log("APNG Edit initializing...");

        // IO service for string -> nsIURI conversion
        this._ioService = Cc["@mozilla.org/network/io-service;1"].getService();
        this._ioService.QueryInterface(Ci.nsIIOService);

        this.ani = document.getElementById("apng");

        this._tmpCanvas = document.getElementById("tmpCanvas");
        this._lightbox = document.getElementById("lightbox");
        // See http://lxr.mozilla.org/seamonkey/source/layout/xul/base/public/nsIScrollBoxObject.idl
        this._lightboxScroller = this._lightbox.boxObject.QueryInterface(Ci.nsIScrollBoxObject);

        this._lightbox.lastChild.framenum = 1;

        this._slowStatusLabel = document.getElementById('slow-status-label');
        this._slowStatusMeter = document.getElementById('slow-status-meter');
    },


    /*
     * addNewFrame
     *
     * Adds a new frame to the lightbox. An optional index may be specified.
     */
    addNewFrame : function (index) {
        var startIndex;

        // Includes the ever-present trailing "new" frame
        var numFrames = this._lightbox.childNodes.length;
        this.log("Adding new frame at index=" + index + " (" + numFrames + " current frames)");
        if (index == undefined)
            index = numFrames;

        var newframe = document.createElement("vbox");
        newframe.setAttribute("newframe", "true");
        newframe.setAttribute("class", "apng_frame");

        if (index < numFrames) {
            // Insert between element[index-1] and element[index]
            var refNode = this._lightbox.childNodes[index];
            this._lightbox.insertBefore(newframe, refNode);
        } else {
            // Insert new frame at end
            this._lightbox.appendChild(newframe);
            if (!this._scrolling) {
                this._scrolling = true;
                window.setTimeout("APNGedit.smoothScrollToEnd();", 0);
            }
        }


        // Renumber following frames, as well as the previous
        // frame (which might be "New Frame" at the moment.
        for (var i = index - 1; i <= numFrames; i++) {
            if (i == -1) { continue; }
            this._lightbox.childNodes[i].framenum = i + 1;
        }

        return newframe;
    },


    /*
     * closeFrame
     *
     * Delete a frame.
     */
    closeFrame : function (frame) {

        var closedFramenum = frame.framenum;
        this._lightbox.removeChild(frame);
        // |frame|'s binding is gone now, so don't access frame.blah

        var numFrames = this._lightbox.childNodes.length;

        // Renumber all the frames (brute force)
        for (var i = closedFramenum - 1; i < numFrames; i++) {
            this._lightbox.childNodes[i].framenum = i + 1;
        }

        this.updateAnimation();

        if (numFrames == 1) {
            this.aniInfo.sizeInitialized = false;
            this.aniInfo.width  = 0;
            this.aniInfo.height = 0;

            var uiWidth  = document.getElementById("ani_width");
            var uiHeight = document.getElementById("ani_height");
            uiWidth.setAttribute("min", "0");
            uiHeight.setAttribute("min", "0");
            uiWidth.value  = 0;
            uiHeight.value = 0;

            this.ani.src = "";
            APNGui.updateAnimationInfobox();
        }
    },


    /*
     * smoothScrollToEnd
     *
     * When adding a new frame, scroll the lightbox to make sure
     * it's visible. This is done as an animation so that things
     * don't suddenly jump around.
     */
    _scrolling : false,
    smoothScrollToEnd : function () {

        var outval = {};
        this._lightboxScroller.getPosition(outval, {});
        var currentPos = outval.value;
        this._lightboxScroller.getScrolledSize(outval, {});
        var totalWidth = outval.value;
        var viewWidth = this._lightbox.boxObject.width;

        var toScroll = totalWidth - viewWidth - currentPos;
        if (toScroll > 10)
            toScroll = toScroll / 3;

        this._lightboxScroller.scrollBy(toScroll, 0); // Don't need to scroll up/down

        if (toScroll > 0)
            window.setTimeout("APNGedit.smoothScrollToEnd();", 10);
        else
            this._scrolling = false;
    },


    /*
     * setAutoPreview
     *
     * Control the automatic (re)encoding of the APNG whenever something
     * changes. Disabling this is useful when doing batch changes, and you
     * don't want to wait for the encoder after each change.
     */
    setAutoPreview : function (enabled) {
        this.autoPreview = enabled;
        this.log("Auto preview " + (enabled ? "enabled" : "disabled"));

        document.getElementById("autopreview").checked = enabled;

        if (enabled && this.stalePreview)
            this.updateAnimation();
    },


    /*
     * setAnimationSize
     *
     * Set the dimensions of the animation.
     */
    setAnimationSize : function (width, height) {
        var uiWidth  = document.getElementById("ani_width");
        var uiHeight = document.getElementById("ani_height");

        if (!this.aniInfo.sizeInitialized) {
            this.aniInfo.sizeInitialized = true;
            // Set the minimum value to 1 now. If we do this in the XUL, it
            // shows as 1x1 at startup, which seems weird.
            uiWidth.setAttribute("min", "1");
            uiHeight.setAttribute("min", "1");
            if (uiWidth.value < 1)
                uiWidth.value = 1;
            if (uiHeight.value < 1)
                uiHeight.value = 1;
        }

        var oldWidth  = this.aniInfo.width;
        var oldHeight = this.aniInfo.height;

        if (!width && !height) {
            // Being called by UI, just get the values
            width = uiWidth.value;
            height = uiHeight.value;
        } else {
            // update UI
            uiWidth.value = width;
            uiHeight.value = height;
        }

        // Ignore unchanged values
        if (width == oldWidth && height == oldHeight)
            return;

        this.log("Setting animation size to " + width + "x" + height);
        this.aniInfo.width  = width;
        this.aniInfo.height = height;
        
        var numFrames = this._lightbox.childNodes.length - 1;

        var oldValue = this.autoPreview;
        this.autoPreview = false;
        // Ensure all frames have the right size (brute force)
        for (var i = 0; i < numFrames; i++) {
            this._lightbox.childNodes[i].animationSizeUpdate(width, height);
        }
        this.setAutoPreview(oldValue);
    },


    /*
     * updateAnimation
     *
     * Encodes the animation. If autopreview is disabled, it simply marks the
     * animation status as dirty and waits for autopreview to be re-enabled.
     */
    updateAnimation : function () {

        if (!this.autoPreview) {
            this.stalePreview = true;
            this.ani.src = "";
            return;
        }

        this.stalePreview = false;

        var numFrames = this._lightbox.childNodes.length - 1;
        if (numFrames == 0) {
            // No frames now. Just reset the animation.
            this.ani.src = "";
            return;
        }

        // If there are still frames loading, bail out now.
        for (var i = 0; i < numFrames; i++) {
            var frame = this._lightbox.childNodes[i];
            if (frame.loading) {
                this.log("Deferring animation update, frame " + (i+1) + " is still loading.");
                // Shouldn't need to set a stale flag, as the last frame to
                // load should cause the animation to refresh itself.
                return;
            }
        }

        this.log("Beginning encode of " + numFrames + " animation frames.");

        if (this._encoding)
            alert("WTF? Attempted to encode APNG before previous encode finished.");
        else {
            this._encoding = true;
            document.getElementById('main-content').setAttribute("busy", "true");
            window.setTimeout(this.asyncEncode, 5, null, numFrames, 0);
            // If we're getting here as a result of an animation script, the
            // caller already started a timeout.
            if (!this._slowID)
                this._slowID = window.setTimeout(this.showSlowUI, 1500);
        }
    },


    /*
     * do_render_frame
     *
     */
    do_render_frame : function (frameNum) {
        var failed = false;
        var numFrames = APNGedit._lightbox.childNodes.length - 1;

        APNGedit.log("Rendering frame " + frameNum);
        APNGedit._slowStatusLabel.value = "Rendering frame " + (frameNum+1) + " of " + numFrames + "...";
        APNGedit._slowStatusMeter.value = Math.round(frameNum / numFrames * 100);

        try {
            var frame = APNGedit._lightbox.childNodes[frameNum];
            frame.useCanvas = true;
            renderFrame(frame, frameNum);
            frame.loading = false;
            frame.updatePreview();
        } catch (e) {
            var msg = "Script failed to render frame " + frameNum + " (" + e + ")";
            APNGedit.log(msg);
            alert(msg);
            failed = true;
        }


        frameNum++;
        if (frameNum == numFrames || failed) {
            // Force an update, which will also remove the busy UI if needed
            // (Unless the render failed, in which case we need to clear the
            // slow UI here)
            APNGedit.updateAnimation();
            if (failed)
                APNGedit.removeSlowUI();

            // We're done.
            APNGedit.setAutoPreview(APNGui.oldAutoPreview);
            return;
        }

        setTimeout(APNGedit.do_render_frame, 5, frameNum);
    },


    /*
     * UI support for long running operations...
     */
    _encoding : false,

    showSlowUI : function () {
        APNGedit._slowID = null;
        document.getElementById('slow-status-overlay').style.display = "";
    },

    removeSlowUI : function () {
        document.getElementById('main-content').setAttribute("busy", "false");
        document.getElementById('slow-status-overlay').style.display = "none";

        window.clearTimeout(APNGedit._slowID);
        APNGedit._slowID = null;
    },

    asyncEncode : function (encoder, numFrames, frameNum) {

        // Note: |this| isn't |APNGedit| here, it's the window. 

        const format = Ci.imgIEncoder.INPUT_FORMAT_RGBA;
        const pixelBytes = 4;
        var options = "";


        if (frameNum == 0) {

            encoder = Cc["@mozilla.org/image/encoder;2?type=image/png"].createInstance();
            encoder.QueryInterface(Ci.imgIEncoder);

            options += ";frames=" + numFrames;
            options += ";skipfirstframe=" + (APNGedit.aniInfo.skipFirstFrame ? "yes" : "no");
            options += ";plays=" + APNGedit.aniInfo.playCount;

            APNGedit.log("startEncode: " + APNGedit.aniInfo.width + "x" + APNGedit.aniInfo.height + " options: " + options);
            encoder.startImageEncode(APNGedit.aniInfo.width, APNGedit.aniInfo.height, format, options);

        } else if (frameNum > numFrames) {
            encoder.endImageEncode();

            var dataURL = APNGedit.getAnimationDataURL(encoder);
            var ani = APNGedit.ani;

            //ani.width  = APNGedit.aniInfo.width;
            //ani.height = APNGedit.aniInfo.height;
            ani.src = dataURL;

            APNGui.updateAnimationInfobox();

            APNGedit.removeSlowUI();
            APNGedit._encoding = false;
            return;

        } else {
            var frame = APNGedit._lightbox.childNodes[frameNum - 1];
            var firstFrame = (frameNum == 1);
            var img = frame.img;

            options = "";
            options += ";delay=" + frame.delay;
            options += ";dispose=" + frame.dispose;
            options += ";blend="   + frame.blend;
            if (!firstFrame) {
                if (frame.offset_x > 0) { options += ";xoffset=" + frame.offset_x; }
                if (frame.offset_y > 0) { options += ";yoffset=" + frame.offset_y; }
            }

            APNGedit.log("addFrame " + frame.framenum + ": " + frame.pixelsW + "x" + frame.pixelsH +
                     " options: " + options + " pixels length: " + frame.pixels.length);

            encoder.addImageFrame(frame.pixels, frame.pixels.length,
                                  frame.pixelsW, frame.pixelsH, frame.pixelsW * pixelBytes, format, options);
        }

        // Update UI now, so that it's shown while we're doing the actual work later.
        if (frameNum++ < numFrames) {
            APNGedit._slowStatusLabel.value = "Encoding frame " + frameNum + " of " + numFrames + "...";
            APNGedit._slowStatusMeter.value = Math.round(frameNum / numFrames * 100);
        } else {
            APNGedit._slowStatusLabel.value = "Updating animation...";
            APNGedit._slowStatusMeter.value = 100;
        }

        // Wait 10ms and then encode the next frame
        window.setTimeout(APNGedit.asyncEncode, 10, encoder, numFrames, frameNum);
    },


    /*
     * onDragOver / onDragDrop callbacks...
     */
    dndAddNewFrame : {
        onDragOver : function (event, flavour, session) {
            //APNGedit.log("onDragOver " + flavour);
        },

        canHandleMultipleItems: true,

        onDrop : function (event, dropData, session) {
            APNGedit.log("APNGedit.onDrop invoked");

            var dataList = dropData.dataList;
            var droppedItems = [];

            for (var i = 0; i < dataList.length; i++) {
                var imgURL;
                var item = dataList[i].first;

                if (item.flavour.contentType == "application/x-moz-file") {
                    APNGedit.log("onDrop got file");
                    var file = item.data.QueryInterface(Ci.nsIFile);
                    var sourceURI = APNGedit._ioService.newFileURI(file);
                    imgURL = sourceURI.spec;
                } else if (item.flavour.contentType == "text/x-moz-url") {
                    APNGedit.log("onDrop got x-moz-url");
                    imgURL = item.data;
                } else {
                    APNGedit.log("onDrop[" + i + "]: unknown drop type: " + item.flavour.contentType);
                }

                APNGedit.log("onDrop[" + i + "]: " + imgURL);
                droppedItems[i] = imgURL;
            }

            // Sort by name, because the items seem to come in random order
            // sometimes...
            droppedItems.sort();

            for (i = 0; i < droppedItems.length; i++) {
                // Just set the source of the last frame, which should always be
                // the trailing "new" frame. It will append a new frame.
                APNGedit._lightbox.lastChild.imgsrc = droppedItems[i];
            }
        },

        getSupportedFlavours : function () {
            var flavours = new FlavourSet();
            flavours.appendFlavour("application/x-moz-file", "nsIFile");
            flavours.appendFlavour("text/x-moz-url");
            //flavours.appendFlavour("text/unicode");
            return flavours;
        }
    },


    /*
     * saveAnimation
     *
     * Save the APNG to a file.
     */
    saveAnimation : function (destFile) {
        var animationDataURL = document.getElementById("apng").src;

        // convert string filepath to an nsIFile
        var file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
        file.initWithPath(destFile);

        var source = this._ioService.newURI(animationDataURL, "UTF8", null);
        var target = this._ioService.newFileURI(file)
  
        // prepare to save the data
        var persist = Cc["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Ci.nsIWebBrowserPersist);
  
        persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
        persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;
  
        // displays a download dialog (remove these 3 lines for silent download)
        //var xfer = Cc["@mozilla.org/transfer;1"].createInstance(Ci.nsITransfer);
        //xfer.init(source, target, "", null, null, null, persist);
        //persist.progressListener = xfer;
  
        // save the data to the file
        persist.saveURI(source, null, null, null, null, file);
    },


    /*
     * getAnimationDataURL
     *
     * Converts the APNG encoder results to a data: URL.
     */
    getAnimationDataURL : function (encoder, mimetype) {
        if (!mimetype) { mimetype = "image/png"; }
            var rawStream = encoder.QueryInterface(Ci.nsIInputStream);

        var stream = Cc["@mozilla.org/binaryinputstream;1"].createInstance();
        stream.QueryInterface(Ci.nsIBinaryInputStream);

        stream.setInputStream(rawStream);

        var bytes = stream.readByteArray(stream.available()); // returns int[]

        return "data:" + mimetype + ";base64," + btoa(String.fromCharCode.apply(null, bytes));
    }
};
